/*
 * Copyright (c) 2003-2005 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nemerle.Collections;

using Nemerle.Compiler.Typedtree;
using Nemerle.Compiler.CompileTypedMethod;

using System.Reflection;

namespace Nemerle.Compiler {

module CompileTypedTree
{
  internal make_ref_type (t : System.Type) : System.Type
  {
#if _NET_2_0
    t.MakeByRefType ()
#else  
    def ty = t.Module.GetType (t.FullName + "&");
    assert (ty != null);
    ty
#endif
  }
  
  internal class CTX
  {
    public mutable current_fun : Fun_header;
    public mutable current_fun_or_loop : Fun_header;
    public mutable current_closure : option [CG_val];
    // FIXME change this to NemerleMap [LocalValue, _]
    public mutable closure_fields : SystemMap [int, FieldInfo];
    public mutable closure_ptrs : SystemMap [int, FieldInfo];
    public mutable locals : SystemMap [int, CG_val];
    public mutable current_type : TypeBuilder;
    public mutable current_loop : option [CExpr.Loop];

    /** Copy constructor */
    public this (copied : CTX) {
      current_fun = copied.current_fun;
      current_closure = copied.current_closure;
      closure_fields = copied.closure_fields;
      closure_ptrs = copied.closure_ptrs;
      locals = copied.locals;
      current_type = copied.current_type;
      current_loop = copied.current_loop;
      current_fun_or_loop = copied.current_fun_or_loop
    }

    /** Empty context */
    public this (ti : TypeBuilder) {
      current_closure = None ();
      closure_fields = SystemMap ();
      closure_ptrs = SystemMap ();
      locals = SystemMap ();
      current_type = ti;
      current_loop = None ();
    }

    public CurrentSystemType : System.Type
    {
      get { current_type.GetTypeBuilder (); }
    }
  }

  internal store_local (ctx : CTX, d : LocalValue, val : CExpr) : CTX * CG_val
  {
    def store_in_closure ()
    {
      match (ctx.current_closure) {
        | Some (clo) =>
          def rf = CExpr.FieldRef (CExpr.Ref (clo), ctx.closure_fields.Get (d.Id));
          CExpr.Assign (rf, val)
        | None =>
          Message.Warning ("no local closure for " + d.Name);
          Util.ice ()
      }
    }
    def var_ty = d.Type.Fix ().GetSystemType ();
    def val =
      match (val) {
        | CExpr.Nop => val
        | _ when val.SystemType.IsValueType && !var_ty.IsValueType => 
          CExpr.Cast (val, var_ty, true)
        | _ => val
      };
    // Message.Debug ($ "store_local: $d $(d.InClosure)");
    if (d.InClosure) {
      def assign = store_in_closure ();
      def cgv = CG_val (loc = d.loc, 
                        name = "dummy",
                        ty = SystemType.Void, 
                        val = assign);
      (CTX (ctx), cgv)
    } 
    else {
      def cgv = CG_val (loc = d.loc, 
                        name = d.Name, 
                        ty = var_ty,
                        val = val);
      def ctx' = CTX (ctx);
      ctx'.locals = ctx.locals.Add (d.Id, cgv);
      (ctx', cgv)
    }
  }

  /** get code fetching closure of given function */
  internal closure_ref (ctx : CTX, fn : Fun_header) : CExpr
  {
    if (ctx.current_fun.id == fn.id)
      CExpr.Ref (Option.UnSome (ctx.current_closure))
    else {
      assert (ctx.closure_ptrs.Member (fn.id),
              $ "broken closure ptr $(fn.name) inside $(ctx.current_fun.name)");
      CExpr.FieldRef (CExpr.This (ctx.CurrentSystemType), ctx.closure_ptrs.Get (fn.id))
    }
  }            

  internal box_value (body : CExpr) : CExpr 
  {
#if _GENERICS
    if (body.SystemType.IsGenericParameter || body.SystemType.IsValueType || SystemType.Void.Equals (body.SystemType))
#else  
    if (body.SystemType.IsValueType || SystemType.Void.Equals (body.SystemType))
#endif
      {
          //Message.Debug (body.ToString ());    
      CExpr.Cast (body, SystemType.Object, true)
      }
    else
      body
  }

  internal unbox_value (body : CExpr, ty : System.Type) : CExpr 
  {
    if (SystemType.Void.Equals (ty))
      CExpr.Ignore (body)
    else
      CExpr.Cast (body, ty, true)
  }

  internal set_parm_info (fp : Fun_parm, v : CG_val) : CG_val
  {
    v.parm_info = CG_parm_info (parm_kind = fp.kind, 
                                modifiers = fp.modifiers);
    v
  }

  internal set_parm_info (v : CG_val) : CG_val
  {
    v.parm_info = CG_parm_info (parm_kind = ParmKind.Normal, 
                                modifiers = Modifiers.Empty);
    v
  }

  global_ref (ctx : CTX, m : IMember) : CExpr {
    match (m.GetKind ()) {
      | MemberKind.Type 
      | MemberKind.Event => assert (false)

      | MemberKind.Field (f) =>
        assert (!f.IsLiteral, "reference to constant field");
        if (f.IsVolatile)
          CExpr.VolatileGlobalRef (f.GetFieldInfo ())
        else
          CExpr.GlobalRef (f.GetFieldInfo ())

      | MemberKind.Property (f) =>
        CExpr.GlobalRef (f.GetPropertyInfo ())
        
      | MemberKind.Method (f) =>
        CExpr.Call (CExpr.ConstructorRef (CompileTypedMethod.get_lambda_proxy (ctx, f)), [])
    }
  }

  internal local_ref (ctx : CTX, d : LocalValue, store : bool) : CExpr
  {
    // fetch value from closure in which it resides
    if (d.InClosure) {
      // special case then we are in global fun and we are asked for 'this'
      if (!store && ctx.current_fun.decl == null && (d.ValKind is LocalValue.Kind.ClosurisedThisPointer))
          CExpr.This (ctx.CurrentSystemType)
      else {
        CExpr.FieldRef (closure_ref (ctx, d.DefinedIn), 
                        ctx.closure_fields.Get (d.Id))
      }
    }
    else if (ctx.locals.Member (d.Id)) {
      match (d.ValKind) {
        | LocalValue.Kind.FunParm (kind) 
          when kind == ParmKind.Out || kind == ParmKind.Ref =>
          def ty = make_ref_type (d.Type.Fix ().GetSystemType ());
          CExpr.Indirect (CExpr.Ref (ty, ctx.locals.Get (d.Id)))
          
        | _ =>
          CExpr.Ref (ctx.locals.Get (d.Id))
      }
    } else {
      match (d.ValKind) {
        | LocalValue.Kind.ClosurisedThisPointer => CExpr.This (ctx.CurrentSystemType)
        | _ =>
          Message.Warning ($ "undef $d = $(d.Id)");
          assert (false)
      }
    }
  }

  internal maybe_cast (_ctx : CTX, needed_type : System.Type, ex : CExpr) : CExpr
  {
#if _GENERICS
//    if (current_type.IsGenericType && !needed_type.IsGenericType &&
//        needed_type.Equals (SystemType.Object)) {
//      Message.Debug ("converting");
//      CExpr.Cast (needed_type, ex, needed_type, true)
//    }       
//    else
      ex
#else
    assert (needed_type != null);
    def current_type = ex.SystemType;
    assert (current_type != null);

    if (needed_type.Equals (SystemType.Void))
      if (current_type.Equals (SystemType.Void))
        ex
      else
        CExpr.Ignore (needed_type, ex)
    else {
/*    
// bug in MS.NET makes it return false for IField[] and IField[]         
      when (needed_type.Equals (current_type) !=
            (needed_type.FullName == current_type.FullName))
        Message.Debug (ex.loc, needed_type.Assembly.ToString () + " " + needed_type.ToString () + " " + 
                               current_type.Assembly.ToString () + " " + current_type.ToString ());
*/                               
      if (needed_type.FullName == current_type.FullName
          || (needed_type.IsAssignableFrom (current_type) && !current_type.IsValueType))
        ex
      else 
        CExpr.Cast (needed_type, ex, needed_type, true)
    }
#endif    
  }

  /**
   * Properly format list of expression before being passed in a function
   * call -- box value types, and add needed downcasts.
   */
  casted_exprs (ctx : CTX, formals : list [CG_val], parms : list [CExpr]) : list [CExpr]
  {
    // Message.Debug ($ "formals : $formals params : $parms");
    def parms = CompileTypedMethod.adjust_tupled_parms (ctx, List.Length (formals), parms);
    List.Map2 (formals, parms, 
      fun (v : CG_val, e) { 
        maybe_cast (ctx, v.ty, e) 
      })
  }

  casted_parms (ctx : CTX, formals : list [Fun_parm], parms : list [Parm]) : list [CExpr]
  {
    def parms = List.Map (parms, fun (p : Parm) { cg_expr (ctx, p.expr) });
    def parms = adjust_tupled_parms (ctx, List.Length (formals), parms);
    List.Map2 (formals, parms, 
      fun (v : Fun_parm, e) { 
        maybe_cast (ctx, v.SystemType, e) 
      })
  }

  emit_loop (ctx : CTX, h : Fun_header, parms : list [Parm]) : CExpr
  {
    def ctx' = CTX (ctx);
    def parms = casted_parms (ctx, h.parms, parms);
    def parm_vals = List.Map2 (h.parms, parms, fun (fp : Fun_parm, ap) {
      def val = CG_val (loc = fp.decl.loc, 
                        name = fp.name,
                        ty = fp.SystemType,
                        val = ap);
      ignore (set_parm_info (fp, val));
      ctx'.locals = ctx'.locals.Add (fp.decl.Id, val);
      val
    });
    ctx'.current_fun_or_loop = h;
    def loop = CExpr.Loop (null, None ());
    ctx'.current_loop = Some (loop);
    loop.body = cg_expr (ctx', (h.body :> FunBody.Typed).expr);
    h.body = null;
    List.FoldRight (parm_vals, loop : CExpr, fun (v, expr) { 
      CExpr.Define (v, expr) 
    })
  }

  with_cached (expr : CExpr, f : CExpr -> CExpr) : CExpr
  {
    match (expr) {
      | CExpr.Ref
      | CExpr.GlobalRef
      | CExpr.VolatileGlobalRef
      | CExpr.ConstructorRef
      | CExpr.MethodAddress
      | CExpr.Opcode
      | CExpr.Literal
      | CExpr.This
      | CExpr.TypeOf 
      | CExpr.Indirect (CExpr.Ref) => f (expr)

      | _ =>
        def v = CG_val (loc = Location_stack.top (),
                        name = Util.tmpname ("cached"),
                        ty = expr.SystemType,
                        val = expr);
        CExpr.Define (v, f (CExpr.Ref (v)))
    }
  }

  cg_call (ctx : CTX, is_tail : bool, ret_type : TyVar, 
           f : TExpr, ps : list [Parm]) : CExpr
  {
    def compile_parm (p : Parm) {
      def ex = cg_expr (ctx, p.expr);
      if (p.kind == ParmKind.Out || p.kind == ParmKind.Ref)
        CExpr.Address (ex)
      else
        ex
    };

    def emit_call (cgm : CG_method, callable) {
      def parms = List.Map (ps, compile_parm);
      def callable =
        match (callable) {
          | CExpr.MethodRef (e, meth, notvirt) when e.SystemType.IsValueType && 
                                         !meth.DeclaringType.IsValueType =>
            CExpr.MethodRef (CExpr.Cast (e, meth.DeclaringType, true), meth, notvirt)
          | x => x
        };
      if (cgm == null) {
        def t = ret_type.Fix ().GetSystemType ();
        if (is_tail)
          CExpr.TailCall (t, callable, parms)
        else
          CExpr.Call (t, callable, parms)
      } else {
        // Message.Debug ("real casted params for " + cgm.name);
        def generate_call (parms) {
          if (is_tail)
            CExpr.TailCall (callable, casted_exprs (ctx, cgm.parms, parms))
          else
            CExpr.Call (callable, casted_exprs (ctx, cgm.parms, parms))
        };
        
        def call =
          match ((parms, cgm.parms)) {
            | ([parm], _ :: _ :: _) =>
              // need to cache it, so casted_parms can untuple it
              with_cached (parm, fun (parm) { generate_call ([parm]) })
            | _ =>
              generate_call (parms)
          };
        maybe_cast (ctx, ret_type.Fix ().GetSystemType (), call)
      }
    };
    def plain_call () {
      def len = List.Length (ps);
      def imeth = get_apply_method (len);
      emit_call (CompileTypedMethod.GetCgMethod (imeth), 
                 CExpr.MethodRef (cg_expr (ctx, f), 
                 imeth.GetMethodInfo (), false))
    };
    def get_instanciated_meth (_fromty : TyVar, member_info)
    {
#if _GENERICS
      def system_ty = _fromty.Fix ().GetSystemType ();
      if (system_ty.HasGenericArguments) {
        Emit.TypeBuilder.GetMethod (system_ty, member_info)
      }
      else member_info
#else                  
      member_info            
#endif                  
    }
    def get_instanciated_cons (_fromty : TyVar, member_info)
    {
#if _GENERICS
      def system_ty = _fromty.Fix ().GetSystemType ();
      Message.Debug (system_ty.ToString ());
      if (system_ty.HasGenericArguments) {
        Emit.TypeBuilder.GetConstructor (system_ty, member_info)
      }
      else member_info
#else                  
      member_info            
#endif                  
    }
 
    match (f) {
      | TExpr.StaticRef (m) =>
        match (m.GetKind ()) {
          | MemberKind.Method (m) =>
            match (m.GetFunKind ()) {
              | FunKind.Constructor =>
                if (m.DeclaringType.IsDelegate)
                  Util.locate (f.loc, CompileTypedMethod.emit_delegate_ctor (ctx, m, ps))
                else {
                  def cons = get_instanciated_cons (ret_type, m.GetConstructorInfo ());
                  emit_call (CompileTypedMethod.GetCgMethod (m), CExpr.ConstructorRef (cons))
                }
                  
              | _ =>
                // GENERICS FIXME: where can I get the instanciated type?
                emit_call (GetCgMethod (m), CExpr.GlobalRef (m.GetMethodInfo ()))                  
            }
          | _ => plain_call ()
        }
      | TExpr.OpCode (name) =>
        emit_call (null, CExpr.Opcode (name))
      
      // check for this invisibly coerced to other type -- encoding for base
      // and force nonvirt then
      | TExpr.MethodRef ((TExpr.LocalRef (ty = ty, decl = decl)) as o, m, false) =>
        def notvirt =
          match ((decl.ValKind, decl.Type.Fix (), ty.Fix ())) {
            | (LocalValue.Kind.ClosurisedThisPointer, 
               MType.Class (ti, _), MType.Class (ti', _)) when !ti.Equals (ti') =>
              true
            | _ => false
          }
        def meth = get_instanciated_meth (ty, m.GetMethodInfo ());
        emit_call (GetCgMethod (m), CExpr.MethodRef (cg_expr (ctx, o), meth, notvirt))
        
      | TExpr.MethodRef (o, m, notvirt) =>
        def meth = get_instanciated_meth (o.Type, m.GetMethodInfo ());         
        emit_call (GetCgMethod (m), CExpr.MethodRef (cg_expr (ctx, o), meth, notvirt))

      | TExpr.Base (m) =>
#if _GENERICS        
        def cons = get_instanciated_cons (ctx.current_type.parent_type,
                                          m.GetConstructorInfo ());
#else                                          
        def cons = m.GetConstructorInfo ();
#endif        
        emit_call (GetCgMethod (m), CExpr.BaseCtorRef (cons))
        
      | TExpr.LocalRef (d) =>
        match (d.ValKind) {
          | LocalValue.Kind.Function (Fun_header where (usage = FunctionUsage.UsedJustOnce) as h, _) =>
            emit_loop (ctx, h, ps);

          | LocalValue.Kind.Function (h, _) =>
            if (h.cg_method.constructor_builder == null)
              // special case, when local method does not have its object                      
              emit_call (h.cg_method, CExpr.GlobalRef (h.cg_method.method_builder))
            else if (h.id == ctx.current_fun.id) {
              assert (h.cg_method.method_builder != null);
              emit_call (h.cg_method,
                         CExpr.MethodRef (CExpr.This (ctx.CurrentSystemType), 
                                          h.cg_method.method_builder, false))
            }
            else plain_call ()

          | _ => plain_call ()
        }
      | _ => plain_call ()
    }
  }

  raw_cg_expr (ctx : CTX, inex : TExpr) : CExpr
  {
    //Message.Debug ("emit");
    def self (e : TExpr) : CExpr { cg_expr (ctx, e) };
    match (inex) {
      | TExpr.LocalRef (d) => local_ref (ctx, d, false)

      | TExpr.StaticRef (d) => global_ref (ctx, d)

      | TExpr.FieldMember (o, f) =>
#if _GENERICS
        def system_ty = o.Type.Fix ().GetSystemType ();
        def field_info = 
          if (system_ty.HasGenericArguments)
            Emit.TypeBuilder.GetField (system_ty, f.GetFieldInfo ())
          else
            f.GetFieldInfo ();
#else                  
        def field_info = f.GetFieldInfo ();
#endif        
        def field_ref =
          if (f.IsVolatile)
            CExpr.VolatileFieldRef (self (o), field_info)
          else
            CExpr.FieldRef (self (o), field_info);
            
        maybe_cast (ctx, inex.ty.Fix ().GetSystemType (), field_ref)

      | TExpr.ImplicitValueTypeCtor (t) =>
        CExpr.InitValueType (t.GetSystemType ())

      | TExpr.MethodRef => Util.ice ("MethodRef")
      | TExpr.PropertyMember => Util.ice ("PropertyMember " + PrettyPrint.SprintTyExpr (None (), inex))
      | TExpr.StaticPropertyRef => Util.ice ("StaticPropertyRef")
      | TExpr.EventMember => Util.ice ("EventMember " + PrettyPrint.SprintTyExpr (None (), inex))
      | TExpr.StaticEventRef => Util.ice ("StaticEventRef")
      | TExpr.ConstantObjectRef (mem) =>
        Message.Warning ("using a constant object reference directly");
        Message.Warning ("  you probably have meant to write `" +
                         mem.DeclaringType.FullName + " ()'");
        def mems =
          mem.DeclaringType.LookupMember ("_N_constant_object_generator");
        match (mems) {
          | [mem] => global_ref (ctx, mem)
          | _ =>
            Message.Error ("  no constant object generator method found");
            CExpr.Nop ()
        }
      
      | TExpr.TypeOf (t) => CExpr.TypeOf (t.Fix ().GetSystemType ())

      | TExpr.SelfTailCall (parms) =>
        def fn = ctx.current_fun_or_loop;
        match (fn.usage) {
          | FunctionUsage.UsedJustOnce =>
            def parms = casted_parms (ctx, fn.parms, parms);
            def assigns = 
              List.Map2 (parms, fn.parms, fun (expr, fp : Fun_parm) {
                (ctx.locals.Get (fp.decl.Id), expr)
              });
            CExpr.Sequence ([CExpr.MultipleAssign (assigns), 
                             CExpr.ContinueLoop (Option.UnSome (ctx.current_loop))])
          | _ =>
            def parms = List.Map (parms, fun (p : Parm) { self (p.expr) });
            // Message.Debug ("self casted params for " + fn.cg_method.name);
            def parms = casted_exprs (ctx, fn.cg_method.parms, parms);
            def t = inex.ty.Fix ().GetSystemType ();
            CExpr.SelfTailCall (t, parms)
        }

      | TExpr.Call (f, ps) =>
        cg_call (ctx, false, inex.ty, f, ps)
       
      | TExpr.TailCall (f, ps) =>
        cg_call (ctx, true, inex.ty, f, ps)

      | TExpr.Assign (e1, e2) =>
        def e1 = self (e1);
        def e2 = self (e2);
        def e2 =
          if (e2.SystemType.IsValueType && !e1.SystemType.IsValueType)
            CExpr.Cast (e2, e1.SystemType, true)
          else
            e2;
        CExpr.Assign (e1, e2)
        
      | TExpr.Array (initializers, dimensions) =>
        def arrtype = inex.ty.Fix ().GetSystemType ();
        def element_type = arrtype.GetElementType ();
        def elt_is_value = element_type.IsValueType;
        def make_initializer (e) {
          def e = self (e);
          if (!elt_is_value && e.SystemType.IsValueType)
            CExpr.Cast (e, element_type, true)
          else
            e
        };
        CExpr.Array (ty = arrtype, 
                     element_type = element_type, 
                     initializers = List.Map (initializers, make_initializer),
                     dimensions = List.Map (dimensions, self))

      | TExpr.ArrayIndexer (ty, obj, args) =>
        CExpr.ArrayAccess (ty.Fix ().GetSystemType (), self (obj), List.Map (args, self))

      | TExpr.DefValIn (d, v, b) =>
        if (d.EverUsed) {
          def (ctx', cgv) = store_local (ctx, d, self (v));
          CExpr.Define (cgv, cg_expr (ctx', b))
        }
        else
          CExpr.Sequence ([CExpr.Ignore (self (v)), cg_expr (ctx, b)])

      | TExpr.DefFunctionsIn (fs, body) =>
        mutable vals = [];
        def store (f : Fun_header, ctx : CTX) : CTX {
          match (f.usage) {
            | FunctionUsage.UsedJustOnce => ctx
            | _ =>
              // here we initialize local functions with new instances of their objects
              match (make_lambda_ctor_call (ctx, f)) {
                | null => ctx // special case when local function has not object
                | ct =>
                  // store instanciated functional object and continue with enriched ctx
                  def (ctx, val) = store_local (ctx, f.decl, ct);
                  vals = val :: vals;
                  ctx
              }
          }
        };
        def ctx = List.FoldLeft (fs, ctx, store);
        foreach (f in fs) {
          match (f.usage) {
            | FunctionUsage.UsedJustOnce => () // we will simply inline its body 
            | _ => emit_lambda_code (ctx, f, (f.body :> FunBody.Typed).expr)
          }
        }
        def mklet (val, acc) { CExpr.Define (val, acc) };        
        List.FoldLeft (vals, cg_expr (ctx, body), mklet)

      | TExpr.Match (e, mcs) =>
        with_cached (cg_expr (ctx, e), 
          fun (e) { 
            MatchingCompiler.Run (inex.ty.Fix ().GetSystemType (), ctx, e, mcs)
          });

      | TExpr.Throw (null) => CExpr.Throw (null)
      | TExpr.Throw (e) => CExpr.Throw (cg_expr (ctx, e))

      | TExpr.TryWith (body, exn, handler) =>
        def body = cg_expr (ctx, body);
        // FIXME: this is broken for exn values to go into closure
        def (ctx', v) = store_local (ctx, exn, CExpr.Nop ());
        def handler = cg_expr (ctx', handler);
        CExpr.TryWith (inex.ty.Fix ().GetSystemType (), body, v, handler);

      | TExpr.TryFinally (e1, e2) =>
        CExpr.TryFinally (cg_expr (ctx, e1), cg_expr (ctx, e2))

      | TExpr.Literal (l) => CExpr.Literal (inex.Type.Fix ().GetSystemType (), l)

      | TExpr.TypeConversion (e, MType.Void, _) =>
        CExpr.Ignore (cg_expr (ctx, e))

      | TExpr.TypeConversion (e, t, ConversionKind.IL (is_checked)) =>
        CExpr.Cast (cg_expr (ctx, e), t.Fix ().GetSystemType (), is_checked)

      | TExpr.TypeConversion (_, t, ConversionKind.MethodCall (meth)) =>
        assert (false, $ "uncatched method call conversion $meth (to $t)");

      | TExpr.TypeConversion (e, t, ConversionKind.IgnoreValue)
      | TExpr.TypeConversion (e, t, ConversionKind.Implicit)
      | TExpr.TypeConversion (e, t, ConversionKind.UpCast) =>
        def e' = cg_expr (ctx, e);
        def t' = t.Fix ().GetSystemType ();
        if (e'.SystemType.IsValueType != t'.IsValueType)
          CExpr.Cast (e', t', true)
        else
          e'

      | TExpr.TypeConversion (e, t, ConversionKind.DownCast) =>
        CExpr.Cast (cg_expr (ctx, e), t.Fix ().GetSystemType (), true)

      | TExpr.TypeConversion (_, _, ConversionKind.Unspecified) =>
        assert (false, "should be handled already")

      | TExpr.TypeConversion (_, _, ConversionKind.Nop)
      | TExpr.TypeConversion (_, _, ConversionKind.Boxing) =>
        assert (false)

      | TExpr.Sequence (e1, e2) =>
        CExpr.Sequence ([cg_expr (ctx, e1), cg_expr (ctx, e2)])

      | TExpr.Tuple (l) =>
        tuple_ctor (List.Map (l, fun (e) { box_value (self (e)) }))

      | TExpr.Base => Util.ice ("stray base")
      | TExpr.OpCode => Util.ice ("stray opcode")

      | TExpr.Delayed => assert (false)
      | TExpr.Error => assert (false)
      | TExpr.This => assert (false, "'this' should have been handled in localrefs and calls");
    }
  }

  /** 
   * Flatten given [Typedtree.TExpr] object into [CExpr].
   *
   * <remarks>
   *   As a side-effect few classes for closures and local functions
   *   can be generated and stored in current type builder.
   * </remarks>
   */
  internal cg_expr (ctx : CTX, ex : TExpr) : CExpr
  {
    Util.locate (ex.loc, raw_cg_expr (ctx, ex))
  }
} // end module
} // end ns
